@@ -12,12 +12,12 @@ import yaml
 from dotenv import load_dotenv, find_dotenv
 from commodore import __git_version__, __version__
 from .catalog import catalog_list
-from .config import Config, Migration
+from .config import Config, Migration, parse_dynamic_facts_from_cli
 from .helpers import clean_working_tree
 from .compile import compile as _compile
 from .component.template import ComponentTemplater
 from .component.compile import compile_component
-from .inventory.render import extract_components
+from .inventory.render import extract_components, extract_packages, extract_parameters
 from .inventory.parameters import InventoryFacts
 from .inventory.lint import LINTERS
 from .login import login, fetch_token
@@ -178,6 +178,26 @@ oidc_client_option = click.option(
     ),
     type=click.Choice([m.value for m in Migration], case_sensitive=False),
 )
+@click.option(
+    "-d",
+    "--dynamic-fact",
+    type=str,
+    metavar="KEY=VALUE",
+    multiple=True,
+    help=(
+        "Fallback dynamic facts to use when compiling a cluster which hasn't "
+        + "reported its dynamic facts yet. Commodore will never use values provided "
+        + "through this parameter if the cluster response from the API has a dynamic "
+        + "facts field. Can be repeated. Commodore expects each fact to be specified "
+        + "as key=value. Nested keys can be provided as `path.to.key`. Commodore will "
+        + "parse values as JSON if they're prefixed by `json:`. If the same key is "
+        + "provided multiple times, the last occurrence overrides the previous values. "
+        + "When providing a value for a key as JSON, previously specified subkeys of "
+        + "that key will be overwritten. Nested keys are ignored if any non-leaf level "
+        + "of the requested key already contains a non-dictionary value. If a value "
+        + "prefixed with `json:` isn't valid JSON, it will be skipped."
+    ),
+)
 @verbosity
 @pass_config
 # pylint: disable=too-many-arguments
@@ -199,6 +219,7 @@ def compile_catalog(
     tenant_repo_revision_override,
     fetch_dependencies,
     migration,
+    dynamic_fact: str,
 ):
     config.update_verbosity(verbose)
     config.api_url = api_url
@@ -213,6 +234,9 @@ def compile_catalog(
     config.migration = migration
     config.oidc_client = oidc_client
     config.oidc_discovery_url = oidc_discovery_url
+    config.fetch_dependencies = fetch_dependencies
+    config.dynamic_facts = parse_dynamic_facts_from_cli(dynamic_fact)
+
     if config.push and (
         config.global_repo_revision_override or config.tenant_repo_revision_override
     ):
@@ -220,8 +244,6 @@ def compile_catalog(
             "Cannot push changes when local global or tenant repo override is specified"
         )
 
-    config.fetch_dependencies = fetch_dependencies
-
     if config.api_token is None and not local:
         try:
             login(config)
@@ -661,18 +683,14 @@ def inventory(config: Config, verbose):
     config.update_verbosity(verbose)
 
 
-@inventory.command(
-    name="components",
-    short_help="Extract component URLs and versions from the inventory",
-)
-@click.option(
+inventory_output_format = click.option(
     "-o",
     "--output-format",
     help="Output format",
     type=click.Choice(["json", "yaml"]),
     default="yaml",
 )
-@click.option(
+inventory_values = click.option(
     "-f",
     "--values",
     help=(
@@ -684,12 +702,64 @@ def inventory(config: Config, verbose):
     multiple=True,
     type=click.Path(exists=True, file_okay=True, dir_okay=False),
 )
-@click.option(
+inventory_allow_missing_classes = click.option(
     " / -A",
     "--allow-missing-classes/--no-allow-missing-classes",
     default=True,
     help="Whether to allow missing classes when rendering the inventory. Defaults to true.",
 )
+
+
+@inventory.command(
+    name="show",
+    short_help="Returns the rendered inventory",
+)
+@inventory_output_format
+@inventory_values
+@inventory_allow_missing_classes
+@click.argument("global-config")
+@click.argument("tenant-config", required=False)
+@verbosity
+@pass_config
+# pylint: disable=too-many-arguments
+def inventory_show(
+    config: Config,
+    verbose,
+    global_config: str,
+    tenant_config: Optional[str],
+    output_format: str,
+    values: Iterable[str],
+    allow_missing_classes: bool,
+):
+    config.update_verbosity(verbose)
+    extra_values = [Path(v) for v in values]
+    try:
+        inv = extract_parameters(
+            config,
+            InventoryFacts(
+                global_config,
+                tenant_config,
+                extra_values,
+                allow_missing_classes,
+                ignore_class_notfound_warning=False,
+            ),
+        )
+    except ValueError as e:
+        raise click.ClickException(f"While rendering inventory: {e}") from e
+
+    if output_format == "json":
+        click.echo(json.dumps(inv))
+    else:
+        click.echo(yaml.safe_dump(inv))
+
+
+@inventory.command(
+    name="components",
+    short_help="Extract component URLs and versions from the inventory",
+)
+@inventory_output_format
+@inventory_values
+@inventory_allow_missing_classes
 @click.argument("global-config")
 @click.argument("tenant-config", required=False)
 @verbosity
@@ -710,7 +780,11 @@ def component_versions(
         components = extract_components(
             config,
             InventoryFacts(
-                global_config, tenant_config, extra_values, allow_missing_classes
+                global_config,
+                tenant_config,
+                extra_values,
+                allow_missing_classes,
+                ignore_class_notfound_warning=False,
             ),
         )
     except ValueError as e:
@@ -722,6 +796,49 @@ def component_versions(
         click.echo(yaml.safe_dump(components))
 
 
+@inventory.command(
+    name="packages",
+    short_help="Extract package URLs and versions from the inventory",
+)
+@inventory_output_format
+@inventory_values
+@inventory_allow_missing_classes
+@click.argument("global-config")
+@click.argument("tenant-config", required=False)
+@verbosity
+@pass_config
+# pylint: disable=too-many-arguments
+def package_versions(
+    config: Config,
+    verbose,
+    global_config: str,
+    tenant_config: Optional[str],
+    output_format: str,
+    values: Iterable[str],
+    allow_missing_classes: bool,
+):
+    config.update_verbosity(verbose)
+    extra_values = [Path(v) for v in values]
+    try:
+        pkgs = extract_packages(
+            config,
+            InventoryFacts(
+                global_config,
+                tenant_config,
+                extra_values,
+                allow_missing_classes,
+                ignore_class_notfound_warning=False,
+            ),
+        )
+    except ValueError as e:
+        raise click.ClickException(f"While extracting packages: {e}") from e
+
+    if output_format == "json":
+        click.echo(json.dumps(pkgs))
+    else:
+        click.echo(yaml.safe_dump(pkgs))
+
+
 @inventory.command(
     name="lint",
     short_help="Lint YAML files for Commodore inventory structures in the provided paths",
@@ -826,4 +943,6 @@ def commodore_fetch_token(
 
 def main():
     load_dotenv(dotenv_path=find_dotenv(usecwd=True))
-    commodore.main(prog_name="commodore", auto_envvar_prefix="COMMODORE")
+    commodore.main(
+        prog_name="commodore", auto_envvar_prefix="COMMODORE", max_content_width=100
+    )
